/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "txtren.h"
#include "graymap.h"
#include "ptrbuf.h"

#include "fsel.h"

#include "fsel-uparr.c"
#include "fsel-dnarr.c"

static u_font_t  *fs_font;
static char      *FS_font_name     = ":default";
static int        FS_font_size     = 12;
static int        FS_vspace        = 3;
static uint32_t   FS_bg            = 0xededed;
static uint32_t   FS_fg            = 0;
static uint32_t   FS_fa_bg         = 0xffffff;
static uint32_t   FS_pa_bg         = 0x7a0068;
static uint32_t   FS_pa_fg         = 0xffffff;

static uint32_t  fm_A= 3, fm_B= 3, fm_C= 3, fm_D= 10,
                 fm_E= 3,          fm_G= 3, fm_H= 3,
                 fm_J= 3;

typedef struct 
{
  char *name;
  char *path;
  int type;
  uint32_t w,h; uint8_t *pixels;
  int y;
} fitem_t;

typedef struct { int x,y,w,h;  } rect_t;

struct xfsel
{
  Window window;
  int w,h;
  ximage_t *image;
  int n_items; 
  fitem_t **items;

  rect_t pa, pi, fa, fi, ua, da;
  int start;
  int end;

  uint8_t *title; uint32_t title_w, title_h;
  int hidden;
  char **sfx;
  void (*click)(void *,file_selector_t*,char*);
  void *arg;

  int selected;
  char *dir_name;
};

static int populate_fs(file_selector_t *fs, char *dn,int hidden);
static void render_fs(file_selector_t *fs);
static void fs_expose(file_selector_t *fs, XExposeEvent *ev);
static void fs_press(file_selector_t *fs, XButtonEvent *ev);
static void compute_areas(file_selector_t *fs);

static void scroll_up(file_selector_t *fs);
static void scroll_down(file_selector_t *fs);
static void click_file(file_selector_t *fs, int y);
static void hilight_item(file_selector_t *fs, int n);

static void fs_destroy(file_selector_t *fs, XEvent *event);


static void free_item(fitem_t *fi)
{
  if (fi->pixels) free(fi->pixels);
  if (fi->path)   free(fi->path);
  if (fi->name)   free(fi->name);
  free(fi);
}

static char *combine_fn(char *dir, char *file)
{
  char *p;
  int i,L;

  if (!strcmp(file,".")) return strdup(dir);
  if (strcmp(file,".."))
  {
     if (!strcmp(dir,"/")) dir= "";
     p= malloc(strlen(dir)+strlen(file)+2);
     sprintf(p,"%s/%s",dir,file);
     return p;
  }
  p= strdup(dir);
  L= strlen(p);
again:
  if (L==0) goto retroot;
  if (p[L-1]=='/') { p[L-1]= 0; L--; goto again; }

  for(i=L-1;i>0;i--)
    if (p[i]=='/') { p[i]=0; return p; }
retroot:
  free(p);
  return strdup("/");
}

static int has_suffix(const char *k, const char *s)
{
  int Lk, Ls,i;
  Lk= strlen(k);
  Ls= strlen(s);
  if (Lk<Ls) return 0;
  for(i=0;i<Ls;i++)
   if (toupper(k[Lk-1-i])!=toupper(s[Ls-1-i]))
     return 0;
  return 1;
}


static fitem_t *mk_fitem(char *dn, char *fn,char **sfx)
{
  fitem_t *fi;
  int Lfn;
  int i;
  struct stat sb;

  fi= malloc(sizeof(*fi));

  fi->path= combine_fn(dn,fn);
  if (stat(fi->path, &sb))      fi->type= 'U';
  else if (S_ISREG(sb.st_mode)) fi->type= 'F';
  else if (S_ISDIR(sb.st_mode)) fi->type= 'D';
  else                          fi->type= 'U';

  if (sfx && fi->type!='D')
  {
    for(i=0;sfx[i];i++) if (has_suffix(fn,sfx[i])) break;
    if (!sfx[i])
    {
       free(fi->path);
       free(fi);
       return NULL;
    }
  }

  Lfn= strlen(fn);
  if (!strcmp(fn,"..")) fi->name= strdup("<UP>");
  else
  {
    fi->name= malloc(Lfn+2);
    strcpy(fi->name, fn);
    if (fi->type=='D')      strcpy(fi->name+Lfn,"/");
    else if (fi->type=='U') strcpy(fi->name+Lfn, "?");
  }

  text_render(fs_font, fi->name, &fi->w, &fi->pixels);
  fi->h= fs_font->height;

  return fi;
}

static void destroy_items(file_selector_t *fs)
{
  int i;
  if (!fs->items) return ;
  for(i=0;i<fs->n_items;i++)
    free_item(fs->items[i]);
  free(fs->items);
}

file_selector_t *file_selector_new
      (Window parent, int x, int y, int w, int h,
       char *dirname, int hidden, char **sfx,
       void (*click)(void *,file_selector_t*,char*), void *arg)
{
  XSetWindowAttributes attribs;
  file_selector_t *fs;


  if (!fs_font) fs_font= font_new(FS_font_name, FS_font_size);
  fs= calloc(1,sizeof(*fs));

  fs->h= h; fs->w= w;
  fs->hidden= hidden;
  fs->sfx= sfx;
  fs->click= click;
  fs->arg= arg;

  attribs.background_pixmap= None;
  fs->window= XCreateWindow(x_display, parent, x,y, w,h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  XSelectInput(x_display, fs->window, ExposureMask | 
                       StructureNotifyMask |
                       ButtonPressMask | ButtonReleaseMask);
  register_event_handler(fs->window, Expose, fs, fs_expose, NULL);
  register_event_handler(fs->window, ButtonPress, fs, fs_press, NULL);
  register_event_handler(fs->window, DestroyNotify, fs, fs_destroy, NULL);

  fs->image= create_image(w,h); 
  fs->selected= -1;
  if (populate_fs(fs, dirname,hidden))
     populate_fs(fs,"/",hidden);
  fs->start= 0;

  compute_areas(fs);
  render_fs(fs);

  XMapWindow(x_display, fs->window);
  return fs;
}

static void fs_destroy(file_selector_t *fs, XEvent *event)
{
  unregister_event_handler(fs->window, Expose, NULL);
  unregister_event_handler(fs->window, ButtonPress, NULL);
  unregister_event_handler(fs->window, DestroyNotify, NULL);
  destroy_image(fs->image);
  destroy_items(fs);

  if (fs->title) free(fs->title);
  if (fs->dir_name) free(fs->dir_name);
  free(fs);
}

void file_selector_free(file_selector_t *fs)
{
  Window window;
  window= fs->window;
  fs_destroy(fs, NULL);
  XDestroyWindow(x_display, window);
}


static int compar_fi(const void *vA, const void *vB)
{
  const fitem_t **fA, **fB, *A, *B;
  const char *types= "DFU";
  const char *tA, *tB;
  fA= (void*) vA; fB= (void*) vB;
  A= *fA; B= *fB;

  if (!strcmp(A->name,"<UP>")) return -1;
  if (!strcmp(B->name,"<UP>")) return 1;
  tA= strchr(types, A->type); if (!tA) tA= types+4;
  tB= strchr(types, B->type); if (!tB) tB= types+4;

  if (tA!=tB) return tA - tB;
  if (A->name[0]=='.' && B->name[0]!='.') return  -1;
  if (A->name[0]!='.' && B->name[0]=='.') return  1;
  return strcmp(A->name, B->name);
}

static int populate_fs(file_selector_t *fs, char *dn,int hidden)
{
  DIR *dir;
  struct dirent *de;
  ptrbuf_t *Bi;
  int y;
  fitem_t **items;
  size_t n_items;
  size_t i;

  dir= opendir(dn);
  if (!dir) return 1;

  Bi= ptrbuf_new(50);
  while((de=readdir(dir)))
  {
     fitem_t *fi;
     if (!strcmp(de->d_name,".")) continue;
     if (strcmp(de->d_name,"..") && !hidden && de->d_name[0]=='.') continue;
     if (!strcmp(dn,"/") && !strcmp(de->d_name,"..")) continue;
     fi= mk_fitem(dn, de->d_name, fs->sfx);
     if (fi) ptrbuf_put(Bi, fi);
  }
  closedir(dir);

  ptrbuf_free(Bi, (void***) &items, &n_items);
  if (n_items==0) return 1;

  fs->selected= -1;

  qsort(items, n_items, sizeof(void*), compar_fi);

  if (fs->title) free(fs->title);
  text_render(fs_font, dn, &fs->title_w, &fs->title);
  fs->title_h= fs_font->height;

  if (fs->dir_name) free(fs->dir_name);
  fs->dir_name= strdup(dn);

  destroy_items(fs);
  fs->items= items;
  fs->n_items= n_items;

  y= 0;
  for(i=0;i<n_items;i++) { items[i]->y= y; y+= items[i]->h + FS_vspace; }

  return 0;
}

static void fs_expose(file_selector_t *fs, XExposeEvent *ev)
{
  put_image(fs->window, ev->x, ev->y,
            fs->image, ev->x, ev->y,
            ev->width, ev->height);
}

static void render_fs(file_selector_t *fs)
{
  uint32_t *iD;
  int i;
  fitem_t *fi;

  iD= fs->image->data;
  fill_rect(iD, fs->w, fs->h, 0,0, fs->w, fs->h, FS_bg);
  fill_rect(iD, fs->w, fs->h, fs->pa.x, fs->pa.y,
                fs->pa.w, fs->pa.h, FS_pa_bg);
  fill_rect(iD, fs->w, fs->h, fs->fa.x, fs->fa.y,
                fs->fa.w, fs->fa.h, FS_fa_bg);

  if (fs->start!=0)
  blend_alpha_image(iD, fs->w, fs->h, fs->ua.x, fs->ua.y,
         (uint32_t*) bi_uparrsmall, bi_uparrsmall_width, bi_uparrsmall_height,
                    0,0, 
                     bi_uparrsmall_width, bi_uparrsmall_height, FS_bg);

  int yoffset;
  yoffset= fs->items[fs->start]->y;

  for(i=fs->start;i<fs->n_items;i++)
  {
    fi= fs->items[i];
    if (fi->y-yoffset >= fs->fi.h) break;

    int W= fi->w;
    if (W>fs->fi.w) W= fs->fi.w;
    int H= fi->h;
    if (H+fi->y-yoffset>= fs->fi.h) H= fs->fi.h - fi->y+yoffset;

    graymap_draw(iD, fs->w, fs->h, fs->fi.x, fi->y+fs->fi.y-yoffset, 
                 fi->pixels, fi->w, fi->h, 0, 0,
                 FS_fg, FS_fa_bg, W, H);

    if (i==fs->selected) hilight_item(fs, i);
  }
  fs->end= i-1;
  if (fs->end<fs->start) fs->end= fs->start;

  if (fs->end!=fs->n_items-1)
    blend_alpha_image(iD, fs->w, fs->h, fs->da.x, fs->da.y,
        (uint32_t*) bi_dnarrsmall, bi_dnarrsmall_width, bi_dnarrsmall_height,
                    0,0, 
                     bi_dnarrsmall_width, bi_dnarrsmall_height, FS_bg);
  
  int tX,tW;

  tX= 0; tW= fs->title_w;
  if (tW>fs->pi.w)
  {
     tW= fs->pi.w;
     tX= tW - fs->pi.w;
  }
  graymap_draw(iD, fs->w, fs->h,  fs->pi.x, fs->pi.y,
                   fs->title, fs->title_w, fs->title_h, tX, 0,
                   FS_pa_fg, FS_pa_bg, tW, fs->title_h); 
}

static void compute_areas(file_selector_t *fs)
{
/********************************
  We have three areas: 
    path area
    file list
    scroll buttons
 *******************************/

  fs->pa.x= fm_B;              // path area
  fs->pa.y= fm_A;
  fs->pa.w= fs->w - 2 *fm_B;
  fs->pa.h= fs_font->height + 2*fm_J;

  fs->pi.x= fs->pa.x + fm_J;   // path area, internal
  fs->pi.y= fs->pa.y + fm_H;
  fs->pi.h= fs_font->height;
  fs->pi.w= fs->pa.w - 2*fm_J;

  fs->fa.x= fm_B + bi_uparrsmall_width + fm_G; // file area
  fs->fa.y= fm_A +  fs->pa.h + fm_C;
  fs->fa.h= fs->h - fs->fa.y - fm_E;
  fs->fa.w= fs->w - fs->fa.x - fm_B;

  fs->fi.x= fs->fa.x + fm_J;            // file area, internal
  fs->fi.y= fs->fa.y + fm_H;
  fs->fi.w= fs->fa.w - 2*fm_J;
  fs->fi.h= fs->fa.h - 2*fm_H;

  fs->ua.x= fm_B;                   // up arrow
  fs->ua.y= (fs->fa.h-bi_uparrsmall_height-bi_dnarrsmall_height-fm_D)/2 
                   + fs->fa.y;
  fs->ua.w= bi_uparrsmall_width;
  fs->ua.h= bi_uparrsmall_height;

  fs->da.x= fs->ua.x;              // down arrow
  fs->da.y= fs->ua.y+ bi_uparrsmall_height + fm_D;
  fs->da.w= bi_dnarrsmall_width;
  fs->da.h= bi_dnarrsmall_height;
}

static int in_rect(XButtonEvent *ev, rect_t *R)
{
  if (ev->x<R->x) return 0;
  if (ev->y<R->y) return 0; 
  if (ev->x>=R->x+R->w) return 0;
  if (ev->y>=R->y+R->h) return 0;
  return 1;
}

static void fs_press(file_selector_t *fs, XButtonEvent *ev)
{
  if (ev->button==Button4) { scroll_up(fs); return ; }
  if (ev->button==Button5) { scroll_down(fs); return ; }
  if (ev->button!=Button1) return ;
  if (in_rect(ev, &fs->ua))  scroll_up(fs);  
  else if (in_rect(ev, &fs->da)) scroll_down(fs);
  else if (in_rect(ev, &fs->fi)) click_file(fs, ev->y);
}

static void scroll_up(file_selector_t *fs)
{
  int i;
  fitem_t *si; int se;

  if (fs->start==0) return ;

  si= fs->items[fs->start];
  se= si->y+si->h;

  for(i=fs->start-1;i>=0;i--)
   if (se - fs->items[i]->y >= fs->fi.h) break;
  if (i==-1) i= 0;
  fs->start= i;
  
  render_fs(fs);
  XClearArea(x_display,fs->window, 0,0, fs->w, fs->h,True);
}

static void scroll_down(file_selector_t *fs)
{
  if (fs->end==fs->n_items-1) return ;
  fs->start= fs->end+1;
  render_fs(fs);
  XClearArea(x_display,fs->window, 0,0, fs->w, fs->h,True);
}

static void hilight_item(file_selector_t *fs, int n)
{
  fitem_t *fi,*si;
  if (n<0 || n>=fs->n_items) return ;
  
  fi= fs->items[n];
  si= fs->items[fs->start];

  int area_e, area_s,h;
 
  h= fi->h;
  area_s= fi->y - si->y;
  area_e= area_s + h;


  if (area_e<=0 || area_s>=fs->fi.h) return ;

  if (area_s<0) { h+= area_s; area_s= 0; }
  if (area_s+h>fs->fi.h) h= fs->fi.h - area_s; 

  if (h<=0) return ;

  xor_rect(fs->image->data, fs->w, fs->h, 
             fs->fi.x, fs->fi.y + area_s, fs->fi.w, h, 0xffffff);

  XClearArea(x_display, fs->window, fs->fi.x, fs->fi.y+area_s,
                      fs->fi.w, h, True);
}

static void click_file(file_selector_t *fs, int y)
{
  int i;
  fitem_t *fi;

  y+= fs->items[fs->start]->y - fs->fi.y;
  for(i=fs->start;i<fs->n_items;i++)
    if (y>=fs->items[i]->y && y<fs->items[i]->y+fs->items[i]->h)
       break;
  if (i==fs->n_items) return ;

  fi= fs->items[i];

  if (fi->type=='D')
  {
    if (populate_fs(fs, fi->path,fs->hidden)) return ;
    fs->start= 0;
    render_fs(fs);
    XClearArea(x_display,fs->window, 0,0, fs->w, fs->h,True);
    return ;
  }

  if (i==fs->selected) return ;

  if (fs->selected!=-1) hilight_item(fs,fs->selected);
  fs->selected= i;
  hilight_item(fs,fs->selected);

  if (fs->click)
   (*fs->click)(fs->arg, fs, fs->items[i]->path);
}

char *file_selector_directory(file_selector_t *fs)
{
  return fs->dir_name;
}
