/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <png.h>
#include <errno.h>
#include "pngio.h"

#define SIG_BYTES 8

static char *open_png_file(char *fn, FILE **fp)
{
  FILE *f;
  char *errstr;
  char ebu[2000];
  unsigned char header[SIG_BYTES];

  f = fopen (fn, "rb");
  if (!f)
  {   
    snprintf(ebu,sizeof(ebu),"can not open file %s: %s", fn, strerror (errno));
    return strdup(ebu);
  }
  if (fread(header, 1, SIG_BYTES, f)!=SIG_BYTES)
  {
    snprintf(ebu,sizeof(ebu),"can not read signature bytes from file %s", fn);
    errstr= strdup(ebu);
    goto err_ret_close;
  }

  if (png_sig_cmp (header, 0, SIG_BYTES))
  {
    snprintf(ebu,sizeof(ebu),"file %s is not a PNG file.", fn);
    errstr= strdup(ebu);
    goto err_ret_close;
  }
  *fp= f;
  return NULL;
err_ret_close:
  fclose(f);
  return errstr;
}

// FIXME: this function leaks memory from data and rows if there
// is an error reading the image.
char *read_png_image
  (char *fn, unsigned int *Rw, unsigned int *Rh, unsigned int **Rdata)
{
  char *errstr;
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  png_infop end_info;
  unsigned int *data, **rows;
  png_uint_32 ww, hh, i;
  int bit_depth, color_type, interlace_type;

  if ((errstr=open_png_file(fn,&fp))) return errstr;

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);

  if (!png_ptr)
    {
      errstr =
	strdup ("internal error. libpng: can not allocate PNG struct.");
      goto err_ret_close;
    }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
    {
      errstr = strdup
	("internal error. libpng: can not allocate first info struct.");
      goto err_ret_destroy0;
    }

  end_info = png_create_info_struct (png_ptr);
  if (!end_info)
    {
      errstr =
	strdup
	("internal error. libpng: can not allocate second info struct.");
      goto err_ret_destroy1;
    }

  if (setjmp (png_jmpbuf (png_ptr)))
    {
      errstr = strdup ("internal error. libpng jumped back.");
      goto err_ret_destroy2;
    }

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  // Get all of the information from the PNG file before the first IDAT (imaged ata chunk).
  png_read_info (png_ptr, info_ptr);

  // Get the goods
  png_get_IHDR (png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

  // tell libpng to strip 16 bit/color files down to 8 bits/color
  png_set_strip_16 (png_ptr);

  // Expand paletted colors into true RGB triplets
  if (color_type == PNG_COLOR_TYPE_PALETTE)
    png_set_expand (png_ptr);

  // Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
    png_set_expand (png_ptr);

  // Expand paletted or RGB images with transparency to full alpha channels
  // so the data will be available as RGBA quartets.
  if (png_get_valid (png_ptr, info_ptr, PNG_INFO_tRNS))
    png_set_expand (png_ptr);

  // Grey images are upgraded to RGB
  if (color_type == PNG_COLOR_TYPE_GRAY
      || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
    png_set_gray_to_rgb (png_ptr);

  // If we dont have true alpha, pad with an alpha channel representing opaque
  png_set_filler (png_ptr, 0xff, PNG_FILLER_AFTER);

  // Turn on interlace handling
  png_set_interlace_handling (png_ptr);

  // Update image info based on transformations
  png_read_update_info (png_ptr, info_ptr);

  data = malloc (ww * hh * sizeof (unsigned int));
  rows = malloc (hh * sizeof (unsigned int *));
  for (i = 0; i < hh; i++)
    rows[i] = data + i * ww;

  png_read_image (png_ptr, (void *) rows);

  // read rest of file, and get additional chunks in info_ptr
  png_read_end (png_ptr, info_ptr);

  // clean up after the read, and free any memory allocated
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);

  /* Now we need to convert from the PNG RGBA format to our uint32_t
     format. The PNG library stores the R byte first, and then G etc.
     We want A at the top 8 bits, then the R, etc. */
  unsigned char *bdata;
  bdata= (unsigned char*) data;
  for(i=0;i<ww*hh;i++)
  {
    unsigned char R,G,B,A; unsigned int P;
    R= bdata[i*4];
    G= bdata[i*4+1];
    B= bdata[i*4+2];
    A= bdata[i*4+3];
    P= (A<<24) | (R<<16) | (G<<8) | B;
    data[i]= P;
  }


  free (rows);
  fclose (fp);

  *Rw = ww;
  *Rh = hh;
  *Rdata = data;

  return 0;

err_ret_destroy0:
  png_destroy_read_struct (&png_ptr, 0, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  png_destroy_read_struct (&png_ptr, &info_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy2:
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}


char *write_png_image
  (char *fn, unsigned int w, unsigned int h, unsigned int *data)
{
  char *errstr;
  FILE *fp;
  char ebu[2000];
  png_structp png_ptr;
  png_infop info_ptr;
  unsigned int **rows;
  png_uint_32 i;

  errstr = 0;
  fp = fopen (fn, "wb");
  if (!fp)
  {
    snprintf(ebu,sizeof(ebu),"can not open file %s: %s", fn, strerror (errno));
    return strdup(ebu);
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (!png_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate PNG struct.");
    goto err_ret_close;
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
  {
    errstr= strdup ("internal error. libpng: can not allocate info struct.");
    goto err_ret_destroy0;
  }

  if (setjmp (png_jmpbuf (png_ptr)))
  {
    errstr = strdup ("internal error. libpng jumped back.");
    goto err_ret_destroy1;
  }

  png_init_io (png_ptr, fp);

  png_set_IHDR (png_ptr, info_ptr,
		w, h, 8,
		PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

  png_write_info (png_ptr, info_ptr);

  rows = malloc (h * sizeof (unsigned int *));
  for (i = 0; i < h; i++)
    rows[i] = data + i * w;

  png_set_filler (png_ptr, 0, PNG_FILLER_AFTER);

  png_write_image (png_ptr, (void *) rows);
  png_write_end (png_ptr, info_ptr);

  png_destroy_write_struct (&png_ptr, &info_ptr);
  free (rows);
  fclose (fp);

  return 0;

err_ret_destroy0:
  png_destroy_write_struct (&png_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  png_destroy_write_struct (&png_ptr, &info_ptr);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}

char *get_png_dimensions (char *fn, unsigned int *Rw, unsigned int *Rh)
{
  char *errstr;
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  png_infop end_info;
  png_uint_32 ww, hh;
  int bit_depth, color_type, interlace_type;

  if ((errstr=open_png_file(fn, &fp))) return errstr;

  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);

//(png_voidp)user_error_ptr, user_error_fn, user_warning_fn);

  if (!png_ptr)
    {
      errstr =
	strdup ("internal error. libpng: can not allocate PNG struct.");
      goto err_ret_close;
    }

  info_ptr = png_create_info_struct (png_ptr);
  if (!info_ptr)
    {
      errstr =
	strdup
	("internal error. libpng: can not allocate first info struct.");
      goto err_ret_destroy0;
    }

  end_info = png_create_info_struct (png_ptr);
  if (!end_info)
    {
      errstr =
	strdup
	("internal error. libpng: can not allocate second info struct.");
      goto err_ret_destroy1;
    }

  if (setjmp (png_jmpbuf (png_ptr)))
    {
      errstr = strdup ("internal error. libpng jumped back.");
      goto err_ret_destroy2;
    }

  png_init_io (png_ptr, fp);
  png_set_sig_bytes (png_ptr, SIG_BYTES);

  // Get all of the information from the PNG file before the first IDAT (imaged ata chunk).
  png_read_info (png_ptr, info_ptr);

  // Get the goods
  png_get_IHDR (png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);

  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);

  fclose (fp);

  *Rw = ww;
  *Rh = hh;

  return 0;

err_ret_destroy0:
  png_destroy_read_struct (&png_ptr, 0, 0);
  fclose (fp);
  return errstr;
err_ret_destroy1:
  png_destroy_read_struct (&png_ptr, &info_ptr, 0);
  fclose (fp);
  return errstr;
err_ret_destroy2:
  png_destroy_read_struct (&png_ptr, &info_ptr, &end_info);
  fclose (fp);
  return errstr;
err_ret_close:
  fclose (fp);
  return errstr;
}
