/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <stdint.h>
#include <syslog.h>
#include "x-basics.h"
#include "x-events.h"



typedef struct 
  { Window window; unsigned int event;
    void *data; void (*func)(void *, XEvent *); }
  event_handler_t;

typedef struct
{
   int key,mod;
   int n_mods;
   int *mods;
   void (*press_func)(void *,XEvent *);
   void (*release_func)(void *,XEvent*);
   void *arg;
} grabkey_t;

static int n_grabkey;
static grabkey_t grabkey[50];
static int grab_exmods[10];
static int grab_nexmods;

static int keyboard_grabbed;
static void (*keyb_press_func)(void *,XEvent *);
static void (*keyb_release_func)(void *,XEvent*);
static void *keyb_arg;

static int n_event_handlers, s_event_handlers;
static event_handler_t *event_handlers;


static event_handler_t *find_event(Window window, unsigned int event);

void register_event_handler(Window window, unsigned int event,
                         void *data, void *func, void**old)
{
  event_handler_t *EV,*CP;
  EV= find_event(window, event);
  if (old) *old= NULL;
  if (EV)
  {
    if (old)
    {
      CP= malloc(sizeof(*CP));
      *CP= *EV;
      *old= CP;
    }
    goto done;
  }

  if (n_event_handlers==s_event_handlers)
  {
    if (s_event_handlers==0)
    {
      s_event_handlers= 20;
      event_handlers= malloc(sizeof(event_handlers[0])*s_event_handlers);
    }
    else
    {
      s_event_handlers+= s_event_handlers/2;
      event_handlers= realloc(event_handlers,
                              sizeof(event_handlers[0])*s_event_handlers);
    }
  }
  EV= event_handlers+n_event_handlers;
  n_event_handlers++;

done:
  EV->window= window;
  EV->event= event;
  EV->data= data;
  EV->func= func;
}

void unregister_all_event_handlers(Window window)
{
  int i;
  for(i=0;i<n_event_handlers;i++)
    if (event_handlers[i].window==window)
    {
      event_handlers[i]= event_handlers[n_event_handlers-1];
      n_event_handlers--; i--;
    }
}

void unregister_event_handler(Window window, unsigned int event, void *old)
{
  event_handler_t *EV,*NE;
  EV= old;
  NE= find_event(window, event);
  if (NE && EV)         *NE= *EV; 
  else if (NE && !EV)   *NE= event_handlers[--n_event_handlers];
  else if (!NE && EV)   register_event_handler(window, event,
                                               EV->data, EV->func,NULL);
  if (EV) free(EV);
}


static event_handler_t *find_event(Window window, unsigned int event)
{
  event_handler_t *EV;
  int i;
  for(i=0;i<n_event_handlers;i++)
  {
    EV= event_handlers+i;
    if (EV->window==window && EV->event==event)
       return EV;
  }
  return NULL;
}


void handle_registered_event(XEvent *event)
{
  event_handler_t *EV;

  if (event->xany.window==x_root_window)
  {
#if 0
    if (event->type==ButtonPress) syslog(LOG_USER|LOG_NOTICE,"root press\n");
    if (event->type==ButtonRelease) syslog(LOG_USER|LOG_NOTICE,"root release\n");
#endif
    if (event->type==KeyPress || event->type==KeyRelease)
    {
      int i;
      if (keyboard_grabbed)
      {
          void (*f)(void*,XEvent *);
          if (event->type==KeyPress) f= keyb_press_func;
                                else f= keyb_release_func;
          if (f) (*f)(keyb_arg,event);
          return ;
      }
      for(i=0;i<n_grabkey;i++)
        if (grabkey[i].key==event->xkey.keycode)
        {
          void (*f)(void*,XEvent *);
          if (event->type==KeyPress) f= grabkey[i].press_func;
                                else f= grabkey[i].release_func;
          if (f) (*f)(grabkey[i].arg,event);
          return ;
        }
    }
  }
                              

  EV= find_event(event->xany.window, event->type);
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void call_event_handler(void *H,XEvent *event)
{
  event_handler_t *EV;
  EV= H;
  if (EV && EV->func)
    (*EV->func)(EV->data, event);
}

void free_event_handler(void *H)
{
  if (H) free(H);
}


static int parse_modname(char *mods, int *Rmod)
{
  int i;
  static const struct { char *name; int modmask; } modnames[]=
  { { "shift", ShiftMask },
    { "lock",  LockMask },
    { "capslock",  LockMask },
    { "control", ControlMask },
    { "mod1", Mod1Mask },
    { "mod2", Mod2Mask },
    { "mod3", Mod3Mask },
    { "mod4", Mod4Mask },
    { "mod5", Mod5Mask },
    { "alt",  Mod1Mask },
    { "numlock", Mod2Mask },
    { NULL, 0 } };
  for(i=0;modnames[i].name;i++)
    if (!strcasecmp(mods,modnames[i].name))
     { *Rmod= modnames[i].modmask; return 0; }
  return 1;
}

static int parse_keyspec(char *keyspec, int *Rmod, int *Rkey)
{
  char *plus;
  int mod;
  KeySym sym;
  KeyCode code;

  mod= 0;
  plus= strchr(keyspec,'+');
  if (plus && plus!=keyspec)
  {
    char *mods;
    mods= strdup(keyspec);
    mods[plus-keyspec]=0;
    if (parse_modname(mods, &mod)) { free(mods); return 1; }
    free(mods);
    keyspec= plus+1;
  }

#if 0
  printf("looking up symbol %s\n",keyspec);
#endif

  sym= XStringToKeysym(keyspec);
  if (sym==NoSymbol) return 1;

  code= XKeysymToKeycode(x_display, sym);
  if (!code) return 1;

  if (Rkey) *Rkey= code;
  if (Rmod) *Rmod= mod;
  return 0;
}

static void make_modifier_list(grabkey_t *gk)
{
  int nemo,i,mod;
 
  mod= gk->mod;

  nemo= grab_nexmods;
  for(i=0;i<nemo;i++)
    if (grab_exmods[i]==mod)
    {
      grab_exmods[i]= grab_exmods[nemo-1];
      grab_exmods[nemo-1]= mod;
      nemo--;
      i--;
    }

  gk->n_mods= (1<<nemo);
  gk->mods= malloc(sizeof(int)*gk->n_mods);

  for(i=0;i<(1<<nemo);i++)
  {
    int M,j;
    M= mod;
    for(j=0;j<nemo;j++)
      if (i&(1<<j)) M|=grab_exmods[j];
    gk->mods[i]= M;
  }
}

int grab_key(char *keyspec, int *Rmod, int *Rkey,
             void *press_func, void *release_func, void *arg)
{
  int mod, key;
  int i;
  grabkey_t *gk;

  if (parse_keyspec(keyspec, &mod, &key)) return 1;
  if (n_grabkey==50) return 1;
  for(i=0;i<n_grabkey;i++)
    if (grabkey[i].key==key && grabkey[i].mod==mod)
       return 1;

  gk= grabkey + n_grabkey;
  n_grabkey++;

  gk->key= key;
  gk->mod= mod;
  gk->press_func= press_func;
  gk->release_func= release_func;
  gk->arg= arg;

  make_modifier_list(gk);
  for(i=0;i<gk->n_mods;i++)
  {
    XGrabKey(x_display, gk->key, gk->mods[i], 
             x_root_window, False, GrabModeAsync, GrabModeAsync);
  }

  if (Rmod) *Rmod= mod;
  if (Rkey) *Rkey= key;
  return 0;
}

void grab_extra_modifiers(char *exmod)
{
  char *copy;
  char *str;
  if (!exmod) exmod= "capslock+numlock";

  copy= str= strdup(exmod);
  while(str[0])
  {
    char *plus,*ne;
    plus= strchr(str,'+');
    if (plus) { *plus= 0; ne= plus+1; }
    else { ne= str+strlen(str); }
    if (!parse_modname(str, grab_exmods+grab_nexmods))
      grab_nexmods++;
    str= ne; 
  }
  free(copy);
}

int grab_keyboard(void *press_func, void *release_func, void *arg)
{
  if (keyboard_grabbed) return 1;
  if (!press_func && !release_func) return 1;
  if (XGrabKeyboard(x_display, x_root_window, False, 
        GrabModeAsync, GrabModeAsync, CurrentTime)!=GrabSuccess) return 1;
  keyboard_grabbed= 1;
  keyb_press_func= press_func;
  keyb_release_func= release_func;
  keyb_arg= arg;
  return 0; 
}

void ungrab_keyboard()
{
  if (!keyboard_grabbed) return ;
  XAllowEvents(x_display, AsyncKeyboard,CurrentTime);
  XUngrabKeyboard(x_display, CurrentTime);
  keyboard_grabbed= 0;
}

